/* $Id: daemon.c,v 1.10 2000/01/15 17:59:55 alinden Exp $ */

/*
*  File             : daemon.c
*  Written by       : alinden@gmx.de
*  Compiler         : egcs on Linux
*  Copyright:       : GPL
*
*  DESCRIPTION: Contains the remaining daemon stuff. Crond jumps into updt() 
*  and calls runproc() periodically to execute programs as necessary.
*/

#include "conf.h"
#include "crond.h"

/* globals as defined in extern.h */

unsigned long uptime=0,starttime=0;
unsigned period=PERIOD;
char *logfile=NULL,sigusr=0;
char logfnam[80];
unsigned qtimeout=DQTMOUT;
int svctrl=-1,svsock=-1,njobs=0;
unsigned cperiod; /* counter for uptime update */
int numqueue=0;   /* entries in queue */
int allowinc=0;   /* allow include for user */
int swaphigh=0;
char sigchld=0,sigusr2=0,tjmp=0;
CONFIG *cblock=NULL,*jblock=NULL,*qblock=NULL;
	
void read_config(void);
void msgout(char *);
char *cpstr(size_t, char *, ...);
void read_fifo(void);
void mail_hmsg(CONFIG *);
void getswap(void);
void run_proc(CONFIG *);
void storequeue(CONFIG *, char);
void runqueue(time_t ntmr);
void flushqueue(void);
int cmd_queued(CONFIG *); 

char *make_msg(char *message)
{
	char *buf;
	int slen; 

	if((buf = strstr(message, "%m")) != NULL && *(buf+2) == 0 && (slen = strlen(message)) < 200)
	{
		buf = strncpy(cpstr(0, NULL), message, slen-2);
		strncpy(buf+slen-2, strerror(errno), 100);
		buf[300] = 0;
		message = buf;
	}
	msgout(message);
	return(message);
}

void log_info(char *message)
{
	openlog("crond", LOG_PID, LOG_CRON);
	syslog(LOG_INFO, "%s", make_msg(message));
	closelog();
}

void log_warn(char *message)
{
	openlog("crond", LOG_PID, LOG_CRON);
	syslog(LOG_WARNING, "%s", make_msg(message));
	closelog();
}

void msgout(char *logmsg)  /* append messages to logfile */
{
	FILE *logfl;
	time_t ltm;
	struct tm *ltime;
	char tmbuf[20];
	
	if(logfile == NULL || (logfl = fopen(logfile, "a")) == NULL)
		return;

	while(time(&ltm) < 0)
		sleep(1);
		
	ltime = localtime(&ltm);
	strftime(tmbuf, 20, "%b  %d %a %H:%M" ,ltime);

	fprintf(logfl, "%s  %s\n", tmbuf, logmsg);
	fclose(logfl);

	return;
}

/* Write/update the uptime file. This contains the time at first startup
   (starttime), \n, relative elapsed time (uptime) in seconds */

int write_upfile(void)
{
	FILE *timefile;

	if((timefile = fopen(TIMEFILE, "w")) != NULL)
	{
		fprintf(timefile, "%lu\n%lu\n", starttime, uptime);
		fclose(timefile);
		return(0);
	}
	log_warn("Uptime file: %m");
	return(-1);
}	

/* cat strings to circulating buffers, max = length of buffer 
   max = 0: only return buffer */

char *cpstr(size_t max, char *fmat, ...)
{
	static char buf[4][1025];
	va_list apl;
	static int bnum=0;

	/* provide 4 buffers */ 

	bnum = ++bnum & 0x03;

	if(! max)
		return(buf[bnum]);
	
	if(max > 1024)
		max = 1024;

	va_start(apl, fmat);
	vsnprintf(buf[bnum], max, fmat, apl);
	buf[bnum][1024] = 0;
	va_end (apl);
	
	return(buf[bnum]);
}

/* return passwd struct for user */

struct passwd *namuid(char *useradr)
{
	struct passwd *pword;

	if((pword = getpwnam(useradr)) == NULL)
	{
		log_warn(cpstr(200, "unknown user: %s", useradr));
		return(NULL);
	}	
	return(pword);	
}

void catf(FILE *infile, FILE *outfile)    /* cat file from infile to outfile */
{
	char buf[512];
	size_t inbytes;

	while((inbytes = fread(buf, 1, 512, infile)))
		fwrite(buf, 1, inbytes, outfile);  	 	 
}   

/* test if m,h,d or m field in CONFIG struct matches current time 
   *tbuf points to bitmap entry (month, hours, day or minute ) in CONFIG 
   ctm = current m,h,d or m time value, return 0 if no match */ 

int chktm(unsigned char *tbuf, int ctm)
{
	int bnum = ctm / 8;

	if ( bnum > 7 )
		return(0);

	/* test if the corresponding bit in the array is set */	

	return((int)(tbuf[bnum] << (unsigned)(ctm-bnum*8)) & 0x80);
}  

/* use chktm() to verify if week or day field match current time 
   *tbuf: current CONFIG struct, ltime: current time struct
   return 0 if no match */

int dwtab(CONFIG *tbuf, struct tm *ltime)
{
	int days,weekdays;

	if(! chktm(tbuf->month, ltime->tm_mon+1))
		return(0); 

	if ( tbuf->wdres == 0x03 )  /* both fields unrestricted */
		return(1);
	
	if ( tbuf->wdres & 0x02 )   /* day field unrestricted */
		days = 0;
	else	
		days = chktm(tbuf->days, ltime->tm_mday);

	if ( tbuf->wdres & 0x01 )   /* week field unrestricted */
		weekdays = 0;
	else	
		weekdays = chktm(tbuf->weekday, ltime->tm_wday);

	return(days | weekdays);
}

/* The following was the most reliable method. No problems with DST timezones
   etc. */ 

time_t runtab(CONFIG *tbuf, time_t ltm)
{
	struct tm *ltime;
	int i,nyear;
	time_t upt,ctm;
	
	ltm += 60;

	if(tbuf->period)
	{
		upt = uptime/60;
		ctm = upt/tbuf->period;

		if(tbuf->period == 1 && ctm)
			ctm--;
			
		ctm = (((ctm*tbuf->period)+tbuf->period-upt)*60)+ltm; 

		return(ctm);
	}
	else
	{
		ltime = localtime(&ltm);
		nyear = ltime->tm_year+1;
		
		if(! dwtab(tbuf, ltime))
		{
NDAY:			ltm += (60*60*21);
			ltime = localtime(&ltm);
		
			for(i = 0; ! dwtab(tbuf, ltime); i++)
			{
				if(i > 500)
					return(-1);

				ltm += (60*60*21);
				ltime = localtime(&ltm);
			}
		
			do
			{
				ltm -= (60*60);
				ltime = localtime(&ltm);
				
			} while(dwtab(tbuf, ltime));

			do
			{
				ltm += 60;
				ltime = localtime(&ltm);
				
			} while(! dwtab(tbuf, ltime));
		} 

		for(i = 0; i < (60*30); i++)
		{
			if(! dwtab(tbuf, ltime))
			{
				if(ltime->tm_year > nyear)
					return(-1);
					
				goto NDAY;
			}
			
			if(ltime->tm_min == 0 && ! chktm(tbuf->hours, ltime->tm_hour))
			{
				ltm += (60*60);
				ltime = localtime(&ltm);
				continue;
			}
				
			if(dwtab(tbuf, ltime) && chktm(tbuf->minutes, ltime->tm_min) && chktm(tbuf->hours, ltime->tm_hour))
				return(ltm);
				
			ltm += 60;
			ltime = localtime(&ltm);
		}
	}	
	return(-1);
}

void restore_sigs(void)
{
	setpgid(0, 0);
	
	signal(SIGTERM, SIG_DFL);
	signal(SIGCHLD, SIG_DFL);
	signal(SIGABRT, SIG_DFL);
	signal(SIGSEGV, SIG_DFL);
	signal(SIGILL, SIG_DFL);
	signal(SIGFPE, SIG_DFL);
	signal(SIGHUP, SIG_DFL);
	signal(SIGALRM, SIG_DFL);
	signal(SIGUSR1, SIG_DFL);
	signal(SIGUSR2, SIG_DFL);
}

/* adjust execution time after time jump */

char runjmp(time_t ltmr, time_t ltm)
{
	CONFIG *tbuf;
	char njmp=0,er=0;

	for(tbuf = cblock; tbuf != NULL; tbuf = tbuf->nblock)
	{ 
		if(ltm >= ltmr+(120*60) || (ltm < ltmr && ltmr - ltm >= (120*60)) || tbuf->period)
		{
			tbuf->nextexec = runtab(tbuf, ltm);
			er = 1;
		}	
		else
			njmp = 1;
			
	}

	if(er)
		log_info("time adjustment performed");

	return(njmp);
}

void getswap(void)  /* detect high swap activity */
{
	static long oldswap=4;
	long newswap=0;
	struct rusage usage;
	static time_t lastmsg=0,lastsave=0;
	time_t ctm = time(NULL);
	
	getrusage(RUSAGE_SELF, &usage);
	newswap = usage.ru_nswap;
	getrusage(RUSAGE_CHILDREN, &usage);
	newswap += usage.ru_nswap;

	if(newswap > oldswap)
	{
		if(logfile != NULL && (! swaphigh || ctm > lastmsg))
		{
			lastmsg = ctm+360;
			msgout("swap activity high");
		}
		swaphigh=1;
	}
	else if(ctm > lastsave)
	{
		if(swaphigh && logfile != NULL)
			msgout("swap activity low");

		swaphigh = 0;
	}

	if(ctm > lastsave)
	{
		oldswap = newswap+4;
		lastsave = ctm+(swaphigh?118:59);
	}	
	return;
}

int q_running(CONFIG *tbuf)  /* running Q job for crontab */
{
	for(tbuf = tbuf->fqblock; tbuf != NULL; tbuf = tbuf->nqblock)
		if(tbuf->pid || cmd_queued(tbuf))
			return(1);

	return(0);
}

void runqproc(time_t ltm, CONFIG *tbuf)   /* exec Q job after another Q job has terminated */
{
	CONFIG *nbuf;

	for(nbuf = tbuf->nqblock; nbuf != NULL; nbuf = nbuf->nqblock)
	{
		if(nbuf->nextexec > -1 && ltm/60 >= nbuf->nextexec/60)
		{
			nbuf->nextexec = runtab(nbuf, ltm);
			run_proc(nbuf);

			break;
		}	
 		else if(nbuf->nextexec == -1 && (ltm/120) % (24*60) == 0)
 			nbuf->nextexec = runtab(nbuf, ltm);
	}
}

void runproc(time_t ltm)   /* exec programs (periodically executed) */
{
	CONFIG *tbuf;

	for(tbuf = jblock; tbuf != NULL; tbuf = tbuf->nblock)
	{
		if(tbuf->htmout && ! tbuf->hsnd && ltm >= tbuf->htmout+tbuf->lastexec)
		{
			mail_hmsg(tbuf);
			tbuf->hsnd = 1;
		}	
	}

	for(tbuf = cblock; tbuf != NULL; tbuf = tbuf->nblock)
	{
		if(tbuf->htmout && tbuf->pid && ! tbuf->hsnd && ltm >= tbuf->htmout+tbuf->lastexec)
		{
			mail_hmsg(tbuf);
			tbuf->hsnd = 1;
		}	
		
		if(tbuf->nextexec > -1 && ltm/60 >= tbuf->nextexec/60)
		{
			if(tjmp || ! tbuf->qopt || ! q_running(tbuf))
			{
				tbuf->nextexec = runtab(tbuf, ltm);
	 			storequeue(tbuf, tjmp);
			}
			else tbuf->qdelay = 1;  /* set delay mark */
	 		
	 	} else if(tbuf->nextexec == -1 && (ltm/120) % (24*60) == 0)
	 		tbuf->nextexec = runtab(tbuf, ltm);
	}
	tjmp = 0;
}

int sig_term(int signo)  /* terminate */
{
	CONFIG *tbuf;
	char fuser=1,*msgterm;
	FILE *rstfile;
	
	write_upfile();
	unlink(PIDFILE);
	unlink(SERVCTRL);

	/* check for jobs to be restarted */
	
	if(signo == SIGTERM)
	{
		rstfile = NULL;
		for(tbuf = cblock; tbuf != NULL; tbuf = tbuf->nblock)
		{
			if(tbuf->stblock)
				fuser = 1;

			if((tbuf->dopt || tbuf->period) && (((! tbuf->ropt) && tbuf->pid) || (tbuf->bopt && cmd_queued(tbuf)) || tbuf->swpdelay || tbuf->qdelay))
			{
				if(rstfile == NULL && (rstfile = fopen(SERVREST, "w")) == NULL)
					break;

				if(fuser)
				{
					fprintf(rstfile, "%%USER %s\n", (tbuf->cmain)?(NCTROOT):(tbuf->username));
					fuser = 0;
				}

				fprintf(rstfile, "%%SUBJ %s %s %s\n", tbuf->username, 
				(tbuf->jobid != NULL)?(tbuf->jobid):("-"), tbuf->cmdstr);
			}
		}
		if(rstfile != NULL)
		{
			fprintf(rstfile, "%%END\n");
			fclose(rstfile);
		}
	}

	log_info((msgterm = cpstr(256, "terminating on signal %d", signo)));
	fprintf(stderr, "crond: %s\n", msgterm);

	exit(0);
}

void joblst(void)
{
	CONFIG *tbuf;
	FILE *msgfile;
	
	if(logfile == NULL || (msgfile = fopen(logfile, "a")) == NULL)
		return;

	fprintf(msgfile, "\n\nreceived USR2 signal, printing joblist\n\n");

	for(tbuf = cblock; tbuf != NULL; tbuf = tbuf->nblock)
	{
		if(tbuf->stblock)
			fprintf(msgfile, "first entry of user\n");
			
		if(tbuf->nuser)
			fprintf(msgfile, "block linked to next user\n");
			
		 	fprintf(msgfile, "period:%lu bopt:%d lopt:%d cmain:%d qtime:%d mailto:%s user:%s cmd:%s\n", tbuf->period, tbuf->bopt, tbuf->lopt, 
		 	        tbuf->cmain, tbuf->qtmout, (tbuf->mailto != NULL)?(tbuf->mailto):("0"), tbuf->username, tbuf->cmdstr);

		if(tbuf->jobid != NULL)
			fprintf(msgfile, "subj: %s\n", tbuf->jobid);
		
		fprintf(msgfile, "-----------------------------------------\n");
	}
	fclose(msgfile);
}

void sig_dexit(int signo) { sleep(1); exit(0); }
void sig_usr2(int signo) { sigusr2 = 1; }
void sig_usr(int signo)  { sigusr = 1;  }
void sig_chld(int signo) { sigchld = 1; }	
		
void updt(void)   /* main routine of running daemon */
{
	time_t ntmr,mins2,mins,ltmr = time(NULL);
	int maxfds,addrlen;
	struct sockaddr_un client;
	char rp;
	fd_set readfds;
	struct timeval tv;

	while(1)
	{
		FD_ZERO(&readfds);
		
		maxfds = svctrl+1;
		
		if(maxfds < 1)
			maxfds = 1;

		while(time(&ntmr) < 1)
			sleep(1);

		mins2 = ntmr / 60;
		mins2 = (mins2*60)+60;

		do
		{
			if(sigchld)
				runqueue(ntmr);

			mins = ntmr / 60;
			mins = (mins*60)+60;
			mins -= ntmr;  /* sleep time */
			
			tv.tv_sec = mins;
			tv.tv_usec = 0;
			
			if(svctrl > -1)
				FD_SET(svctrl, &readfds);

			rp = 0;
	
			/* Awake on events */	

			if(select(maxfds, &readfds, NULL, NULL, &tv) > 0 && FD_ISSET(svctrl, &readfds))
				rp = 1;

			if(sigusr)
			{
				sigusr = 0;
				read_config();
			}	

			if(sigusr2)
			{
				sigusr2 = 0;
				joblst();
			}

			while(time(&ntmr) < 1)
				sleep(1);

			if(ltmr > ntmr || ntmr >= ltmr+120)  /* time jump */
			{
				if(runjmp(ltmr, ntmr) && ntmr > ltmr)
					tjmp = 1;

				if(rp)
					uptime -= 30;	
					
				mins2 = ntmr;
			}
			
			if(rp)  /* socket connection */
			{
				if((svsock = accept(svctrl, (struct sockaddr *)&client, &addrlen)) == -1)
					log_warn("accept call failed: %m");
				else
				{
					fcntl(svsock, F_SETFD, 1);
					read_fifo();
					close(svsock);
					svsock = -1;
				}
			}

		} while(ntmr < mins2);
		
		getswap();
		
		uptime += 60;
		runqueue(ntmr);

		ltmr = ntmr;	 	 
		
		if(--cperiod == 0)
		{
			write_upfile();
			cperiod = period;
		}	
		
		runproc(ntmr);
	}
}
