/* $Id: main.c,v 1.12 1999/11/15 19:40:16 alinden Exp $ */

/*
*  File             : main.c
*  Written by       : alinden@gmx.de
*  Compiler         : egcs on Linux
*  Portability      : Should be POSIX compliant  
*  Copyright:       : GPL
*
*  DESCRIPTION: Startup routines
*/

#include "conf.h"
#include "crond.h"
#include "extern.h"

void read_config(void);
int write_upfile(void);
void log_info(char *);
void log_warn(char *);
int sig_term(void);
void sig_usr(void);
void sig_usr2(void);
void sig_chld(void);
int chktm(char *, int);
void run_proc(CONFIG *);
int dwtab(CONFIG *, struct tm *); 
time_t runtab(CONFIG *, time_t);
char *cpstr(size_t, char *, ...);
int cmd_queued(CONFIG *);
void storequeue(CONFIG *, char);
void updt(void);

void daemon_init(void)
{
	pid_t pid;
	char buf[20];
	int fd,fdval;

	/* Fork a daemon process */

	if((pid = fork()) < 0) 
	{	
		perror("cannot fork");
		exit(1);
	} 

	if(pid)
		exit(0);

	if((fd = open(PIDFILE, O_WRONLY | O_CREAT, 0644)) == -1)
	{
		perror("Cannot open pidfile");
		exit(1);
	}
	
	/* verify if another daemon is running */

	errno = 0;
	if(flock(fd, LOCK_EX | LOCK_NB) == -1)
	{
		if(errno == EWOULDBLOCK)
		{
			fputs("Another daemon is already running\n", stderr);
			exit(1);
		}

		perror("flock error");
	}
	else
	{
		/* Write the PID */ 

		if(ftruncate(fd, 0) == -1)
			perror("ftruncate call failed");
		else
		{
			sprintf(buf, "%ld\n", (long)getpid());
			write(fd, buf, strlen(buf));
		}
	}

	errno = 0;
	if((fdval = fcntl(fd, F_GETFD, 0)) != -1)
	{
		fdval |= FD_CLOEXEC;
		fcntl(fd, F_SETFD, fdval);
	}

	if(errno)
		perror("fcntl error");
	
	umask(022);
	setsid();
}

void init_uptime(void)  /* Read the uptime file */
{
	FILE *timefile;
	
	if((timefile = fopen(TIMEFILE, "r")) != NULL)
	{
		if(fscanf(timefile, "%lu\n%lu", &starttime, &uptime) > 1)
		{
			fclose(timefile);
			return;
		}
		fclose(timefile);
	}		

	/* If necessary,initialize it */

	starttime = time(NULL);
	if(write_upfile() == 0)
	{
		log_info("new counter created");
		return;
	}
	fputs("can't write uptime\n", stderr);
	unlink(PIDFILE);
	exit(1);
}	

void runimd(void)   /* test for immediate execution (D-option used) */
{
	struct stat buf;
	time_t ltm,ctm;
	CONFIG *tbuf;
	int jobs=0;
	
	if(stat(TIMEFILE, &buf))
		return;

	while(time(&ltm) < 0)
		sleep(1);

	if(! buf.st_mtime || ! ltm || buf.st_mtime > ltm)
		return;
	
	if(buf.st_mtime > 60)
		buf.st_mtime -= 60;

	for(tbuf = cblock; tbuf != NULL; tbuf = tbuf->nblock)
	{ 
		if(tbuf->dopt && ! tbuf->pid && ! cmd_queued(tbuf) && 
		  (ctm = runtab(tbuf, buf.st_mtime)) > -1 && ctm <= ltm)
		{
			storequeue(tbuf, 1);
			jobs++;
		}		
	}

	if(jobs)
		log_info(cpstr(100, "executed %d jobs after restart", jobs));
}

CONFIG *fnduser(char *username)  /* find first block of a user */
{
	CONFIG *tbuf;
	char cmain=0;

	if(strcmp(username, NCTROOT) == 0)
		cmain = 1;

	for(tbuf = cblock; tbuf != NULL; tbuf = tbuf->nuser)
	{
		if((cmain && tbuf->cmain) ||
		  (! cmain && ! tbuf->cmain && strcmp(username, tbuf->username) == 0))
			break;
	}
	return(tbuf);
}
	
void resimd(void)   /* test for restart file */
{
	CONFIG *tbuf,*pbuf=NULL;
	FILE *rstfile;
	int jobs=0,cread;
	char *rfield[4];
	char tcmd[10*1025];
	
	if((rstfile = fopen(SERVREST, "r")) == NULL)
		return;

	tcmd[(10*1025)-1]=0;

	while(fgets(tcmd, 10*1024, rstfile) != NULL)
	{
		*(tcmd+strcspn(tcmd, "\n")) = 0;

		if(strncmp(tcmd, "%END", 4) == 0)
			break;

		if((cread=fgetc(rstfile)) == EOF || ungetc(cread, rstfile) != cread)  /* incomplete file */
		{
			log_warn("incomplete restart file");
			break;
		}
		
		if(strncmp(tcmd, "%USER", 5) == 0)
		{
			pbuf = fnduser(tcmd+6);
			continue;
		}	

		if(pbuf == NULL)
			continue;
		
		if((rfield[0] = strtok(tcmd, "\t\n ")) == NULL ||  
		    strcmp(rfield[0], "%SUBJ") ||
		    (rfield[1] = strtok(NULL, "\t\n ")) == NULL ||  /* name */
		    (rfield[2] = strtok(NULL, "\t\n ")) == NULL ||  /* subject */
		    (rfield[3] = strtok(NULL, "\0")) == NULL)       /* command */
			continue;
		
		tbuf = pbuf;
		while(tbuf != NULL)  /* find entry */
		{
			if(strcmp(tbuf->username, rfield[1]) == 0 && (tbuf->jobid == NULL || strcmp(tbuf->jobid, rfield[2]) == 0) &&
			   strcmp(tbuf->cmdstr, rfield[3]) == 0)
				break;

			tbuf = tbuf->nblock;

			if(tbuf != NULL && tbuf->stblock)
				tbuf = NULL;
		}

		if(tbuf == NULL)
			continue;

		storequeue(tbuf, 1);
		jobs++;
	}
	fclose(rstfile);
	
	if(jobs)
		log_info(cpstr(100, "restarted %d interrupted jobs", jobs));
}

int main(int argc, char *argv[])
{
	unsigned bdelay = BDELAY;
	sigset_t sigmask;
	struct sigaction act;
	int i; 

	if(argc > 1 && strcmp(argv[1], "--version") == 0)
	{
		puts(VERSION);
		exit(0);
	}

	for(i = 1; i < argc; i++)
	{
		if( *argv[i] != '-' )
			continue;

		switch(*(argv[i]+1))
		{
			case 'u':
					if(++i < argc) 
						period = atoi(argv[i]);
					break;
			case 'd':
					if(++i < argc)
						bdelay = atoi(argv[i]);
					break;
			case 'i':
					allowinc = 1;
					break;
			case 'l':
					if(++i < argc)
					{
						strncpy(logfnam, argv[i], 80);
						logfnam[79] = 0;
						logfile = logfnam;
					}	
					break;
			case 't':
					if(++i < argc)
						qtimeout = atoi(argv[i]);
					break;
			default:
					fprintf(stderr, "%s%s\n", "Unknown option: ", argv[i]);
		}
	}

	daemon_init();
	init_uptime();

	log_info("restart");
	
	if(bdelay)
	{
		log_info(cpstr(100, "delayed %ds", bdelay));
		sleep(bdelay);
	}	
		
	read_config();

	if(period < 1)
		period = 1;

	cperiod = period;	

	sigfillset (&sigmask);
	sigdelset(&sigmask, SIGTERM);	
	sigdelset(&sigmask, SIGALRM);
	sigdelset(&sigmask, SIGABRT);
	sigdelset(&sigmask, SIGSEGV);
	sigdelset(&sigmask, SIGILL);
	sigdelset(&sigmask, SIGFPE);
	sigdelset(&sigmask, SIGHUP);
	sigdelset(&sigmask, SIGUSR1);
	sigdelset(&sigmask, SIGUSR2);
	sigdelset(&sigmask, SIGCHLD);

	i = 0;

	act.sa_handler = (void *)sig_term;
	sigfillset(&act.sa_mask);
	act.sa_flags = 0;
	if(sigaction(SIGTERM, &act, NULL) < 0 ||
	sigaction(SIGABRT, &act, NULL) < 0 ||
	sigaction(SIGSEGV, &act, NULL) < 0 ||
	sigaction(SIGILL, &act, NULL) < 0 ||
	sigaction(SIGFPE, &act, NULL) < 0)
		i = 1;

	act.sa_handler = (void *)sig_chld;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_NOCLDSTOP | SA_RESTART;
	if(sigaction(SIGCHLD, &act, NULL) < 0)
		i = 1;

	act.sa_handler = (void *)sig_usr2;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_RESTART;	
	if(sigaction(SIGUSR2, &act, NULL) < 0)
		i = 1;

	act.sa_handler = (void *)sig_usr;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_RESTART;
	if(sigaction(SIGUSR1, &act, NULL) < 0 ||
	sigaction(SIGHUP, &act, NULL) < 0 || i)
		log_warn("Can't catch signals");

	sigprocmask(SIG_SETMASK, &sigmask, NULL);
		
	resimd();
	unlink(SERVREST);
	runimd();	
	
	updt();

	exit(0);
}
